from datetime import datetime, timedelta
from enum import Enum
from dataclasses import dataclass

import numpy as np
import pandas as pd

from toargridding.toarstats_constants import STATISTICS_LIST
from toargridding.static_metadata import global_cf_attributes, TOARVariable

date_created = datetime.utcnow().strftime("%Y-%m-dT%H:%M:%SZ")

COORDINATE_VARIABLES = ["latitude", "longitude", "time"]
DATA_VARIABLES = ["mean", "std", "n"]


DataVariables = Enum("DataVariables", DATA_VARIABLES)
Coordinates = Enum("Coordinates", COORDINATE_VARIABLES)
Variables = Enum("Variables", [*DATA_VARIABLES, *COORDINATE_VARIABLES])

TIME_FREQUENCIES = {"daily": "D", "monthly": "M", "yearly": "Y"}  # TODO test


# TODO allow only valid sampling strings
@dataclass
class TimeSample:
    start: datetime
    end: datetime
    sampling: str

    def as_datetime_index(self):
        return pd.period_range(self.start, self.end, freq=self.frequency).to_timestamp()

    @property
    def daterange_option(self):
        end_with_padding = self.end + timedelta(1)
        return f"{self.start.isoformat()},{end_with_padding.isoformat()}"

    @property
    def frequency(self):
        return TIME_FREQUENCIES[self.sampling]

    def as_cf_index(self):
        n_days = (self.end - self.start).days
        return np.arange(n_days + 1)


@dataclass
class Metadata:
    variable: TOARVariable
    statistic: str
    time: TimeSample

    @staticmethod
    def construct(standart_name: str, stat: str, time: TimeSample):
        variable = TOARVariable.get(standart_name)

        if stat not in STATISTICS_LIST:
            raise ValueError(f"invalid statistic: {stat}")

        return Metadata(variable, stat, time)

    def get_id(self):
        return f"{self.variable.name}_{self.statistic}_{self.time.daterange_option}_at_{datetime.now().date().isoformat()}"

    def get_title(self):
        return f"{self.time.sampling} {self.statistic} statistic for {self.variable.standart_name} from {self.time.start} to {self.time.end} aggregated on global grid"

    def get_summary(self):
        return f"{self.time.sampling} {self.statistic} statistic for {self.variable.standart_name} from {self.time.start} to {self.time.end} aggregated on global grid"


@dataclass
class AnalysisRequestResult:
    stations_data: pd.DataFrame
    stations_coords: pd.DataFrame
    metadata: Metadata


def get_global_attributes(metadata: Metadata):
    dynamic_cf_attributes = {
        "id": metadata.get_id(),
        "title": metadata.get_title(),
        "summary": metadata.get_summary(),
        "date_created": date_created,
        "history": f"{date_created}: File created by toargridding package using data from toar-analysis service",
        # "geospatial_bounds": 0,  # for polygonal desrcription
        "geospatial_lat_min": -90,  # TODO read from grid description
        "geospatial_lat_max": 90,
        "geospatial_lon_min": -180,
        "geospatial_lon_max": 180,
        "time_coverage_start": metadata.time.start.isoformat(),
        "time_coverage_end": metadata.time.end.isoformat(),
        # "time_coverage_duration": 0, # TODO insert durations
        # "time_coverage_resolution": 0,
    }
    cf_attributes = dynamic_cf_attributes | global_cf_attributes
    return cf_attributes


def get_cf_metadata(variable: Variables, metadata: Metadata | None):
    match variable.name:
        case Variables.latitude.name:
            cf_metadata = {
                "attributes": {
                    "standard_name": "latitude",
                    "long_name": "latitude in degrees north",
                    "units": "degrees_north",
                    "coverage_type": "coordinate",
                },
                "encoding": {"dtype": "int32"},
            }
        case Variables.longitude.name:
            cf_metadata = {
                "attributes": {
                    "standard_name": "longitude",
                    "long_name": "longitude in degrees east",
                    "units": "degrees_east",
                    "coverage_type": "coordinate",
                },
                "encoding": {"dtype": "int32"},
            }
        case Variables.time.name:
            cf_metadata = {
                "attributes": {
                    "units": f"days since {metadata.time.start.isoformat()}",  # TODO formatting !!
                    "long_name": "time",
                    "standard_name": "time",
                    "calendar": "standard",
                    "coverage_content_type": "coordinate",
                },
                "encoding": {"dtype": "int32"},
            }
        case Variables.mean.name:
            cf_metadata = {
                "attributes": {
                    "units": metadata.variable.units,
                    "long_name": metadata.variable.long_name,
                    "standard_name": metadata.variable.standart_name,
                    "cell_methods": "time: point, latitude: mean, longitude: mean",
                    "coverage_content_type": "physicalMeasurement",
                },
                "encoding": {"dtype": "float32", "_FillValue": np.nan, "zlib": False},
            }
        case Variables.std.name:
            cf_metadata = {
                "attributes": {
                    "units": metadata.variable.units,
                    "long_name": metadata.variable.long_name,
                    # TODO standart name extension vs cell_methods
                    "standard_name": f"{metadata.variable.standart_name} standard_error",
                    "cell_methods": "time: point, latitude: standard_deviation, longitude: standard_deviation",
                    "coverage_content_type": "physicalMeasurement",
                },
                "encoding": {
                    "dtype": "float32",
                    "_FillValue": np.nan,
                    "zlib": False,
                },
            }
        case Variables.n.name:
            cf_metadata = {
                "attributes": {
                    "units": "1",
                    "long_name": f"number of {metadata.variable.long_name} observations in cell",
                    "standard_name": "number_of_observations",
                    "cell_methods": "time: point, latitude: sum, longitude: sum",
                    "coverage_content_type": "auxiliaryInformation",
                },
                "encoding": {
                    "dtype": "int32",
                    "_FillValue": 0,
                    "zlib": False,
                },
            }

    return cf_metadata
